#include "piece.h"
#include "inputspec.h"
#include "configuration.h"

#include <iostream>

using namespace std;


// Directional (x, y) offsets.  0 = right, 1 = down, 2 = left, 3 = up.
// This lets us use (dir + 2) % 4 to get a direction opposite to dir.

const int DX[] = { 1, 0, -1, 0 };
const int DY[] = { 0, 1, 0, -1 };

InputSpec *Piece::spec;

// Make a new piece of a given shape, in a given place.

Piece::Piece(int whichShape, int l, int t)
{
    shapeIndex = whichShape;
    left = l;
    top = t;
}


// Make a new piece copied from an old one.

Piece::Piece(Piece &piece)
{
    *this = piece;
}


// Is this the goal piece, in the goal position?

bool Piece::solved()
{
    return isGoal() && left == spec->goalX && top == spec->goalY;
}


// Is this the goal piece?

bool Piece::isGoal()
{
    return shapeIndex == spec->goalIndex;
}


// Does this piece occupy the spot (x, y)?

bool Piece::occupied(int x, int y)
{
    x -= left;
    y -= top;
    
    if (x < 0 || y < 0 || x >= MAXPIECEDIM || y >= MAXPIECEDIM)
        return false;
        
    return spec->shapes[shapeIndex][x][y];
}


// Can this piece slide in the specified direction, in the given configuration?
// We must make sure that the squares the piece would slide into are vacant.
// Those are the ones that are in the given direction from the squares the
// piece occupies, except for those such squares that the piece itself
// currently occupies.

bool Piece::canSlide(int dir, Configuration *config)
{
    for (int x = left; x <= left + MAXPIECEDIM; ++x)
        for (int y = top; y <= top + MAXPIECEDIM; ++y)
            if (occupied(x, y) && !occupied(x + DX[dir], y + DY[dir]) && config->occupied(x + DX[dir], y + DY[dir]))
                return false;
                
    return true;
}       


// Shift the piece.

void Piece::shift(int dir)
{
    left += DX[dir];
    top += DY[dir];
}


// Print out the piece info.

void Piece::print()
{
    cout << shapeIndex << " " << left << " " << top << endl;
}
    

// Is this piece the same shape as and in the same location as another piece?
// Used for comparing configurations.

bool Piece::sameAs(Piece *piece)
{
    return shapeIndex == piece->shapeIndex && left == piece->left && top == piece->top;
}


// Make a new piece list node.

PieceList::PieceList(Piece *p, PieceList *pl)
{
    piece = p;
    next = pl;
}


// Make a copy of an existing piece list.

PieceList::PieceList(PieceList *pl)
{
    piece = new Piece(*pl->piece);
    next = pl->next ? new PieceList(pl->next) : NULL;
}


// Dispose the contents of the list.

void PieceList::deleteContents()
{
    delete piece;
    
    if (next)
    {
        next->deleteContents();
        delete next;
    }
}


// Print a PieceList.  The list is printed in order; this matches the order
// piece numbers are assigned when printing solutions.

void PieceList::print()
{
    piece->print();

    if (next)
        next->print();  
}


// Find the pieceNumth piece in the list.

Piece *PieceList::index(int pieceNum)
{
    return pieceNum ? next->index(pieceNum - 1) : piece;
}
